/*
 * Renders a simple water surface sim with pleasant Gooch-tone gradients
 */


/* puddles, Copyright (c) 2012 Christopher Oat <chris@chrisoat.com> 
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or
 * implied warranty.
 *
 */



/*
 * TODO: - Fix the auotmip gen for the background texture (ENABLE_BACKGROUND_MIP)
 *       
 */

#define DEFAULTS	"*delay:	10000       \n" \
"*count:        30          \n" \
"*showFPS:      False       \n" \
"*wireframe:    False       \n" \

# include "xlockmore.h"		/* from the xscreensaver distribution */

#ifdef USE_GL /* whole file */

#define DEFAULT_SIMULATION_RESOLUTION   ("1")      // 0 = 256, 1 = 512, 2 = 1024, 3 = 2048
#define DEFAULT_DROPLETS_PER_SECOND     ("16")
#define DEFAULT_GRADIENT_TIME           ("15")
#define DEFAULT_RIPPLE_HEIGHT           ("2")


// You can try changing this if your GPU doesn't support a given format
// 0 = uint8, 1 = uint10, 2 = uint16, 3 = fp16, 4 = fp32
#define SIMULATION_TEXTURE_FORMAT  (3)

// Enabling this crashes older ATI cards, probably because the background texture isn't a power of 2.
#define ENABLE_BACKGROUND_MIP     (0)

#define DROPLET_MAX_AGE           (0.5f)
#define MAX_DROPLET_COUNT         (512)

// Gradient contrast threshold for allowing a top->bottom warm/cool flip. Lower values result in more flips but run the
// risk of some color gradient blends that don't have much vertical contrast
#define CONTRAST_THRESHOLD        (0.5f)

// Enable extra logging and error checking after every GL API call
#define ENABLE_GL_DEBUGGING (0)

// Enable support for dumping frames for screen shots or composing a movie
#define ENABLE_FRAME_DUMPING (0)
#define FRAME_DUMP_LOCATION ("/Users/Chris/Movies/Puddle Movie/Frames")

typedef enum
{
    SIM_RESOLUTION_256x256   = 0,
    SIM_RESOLUTION_512x512   = 1,
    SIM_RESOLUTION_1024x1024 = 2,
    SIM_RESOLUTION_2048x2048 = 3
} SimResEnum;

static SimResEnum g_nSimResolutionEnum;
static int g_nNumDropletsPerSecond;
static float g_fGradientTransitionTime;
static float g_fRippleHeight;

static XrmOptionDescRec opts[] =
{
    {"-dps",        ".puddles.dps",         XrmoptionSepArg, 0},
    {"-colortime",  ".puddles.colortime",   XrmoptionSepArg, 0},
    {"-simres",     ".puddles.simres",      XrmoptionSepArg, 0},
    {"-ripheight",  ".puddles.ripheight",   XrmoptionSepArg, 0},
};

static argtype vars[] =
{
    {&g_nNumDropletsPerSecond,      "dps",          "DPS",          DEFAULT_DROPLETS_PER_SECOND,    t_Int},
    {&g_fGradientTransitionTime,    "colortime",    "ColorTime",    DEFAULT_GRADIENT_TIME,          t_Float},
    {&g_nSimResolutionEnum,         "simres",       "SimRes",       DEFAULT_SIMULATION_RESOLUTION,  t_Int},
    {&g_fRippleHeight,              "ripheight",    "RipHeight",    DEFAULT_RIPPLE_HEIGHT,          t_Float},
};

static OptionStruct desc[] =
{
	{"-dps num",        "droplets per second"},
    {"-colortime num",  "color transition time"},
    {"-simres num",     "sim resolution: 0=256, 1=512, 2=1024, 3=2048"},
    {"-ripheight num",  "ripple height"},
};

ENTRYPOINT ModeSpecOpt puddles_opts =
{sizeof opts / sizeof opts[0], opts, sizeof vars / sizeof vars[0], vars, desc};


// 3 Component vector
typedef struct
{
	union
	{
		float XYZ[3];
		float RGB[3];
		struct
		{
			float r,g,b;
		};
		struct
		{
			float x,y,z;
		};
	};
	
} Vector3;

typedef Vector3 Color3;

// 4 Component vector
typedef struct
{
	union
	{
		float XYZW[4];
		float RGBA[4];
		struct
		{
			float r,g,b,a;
		};
		struct
		{
			float x,y,z,w;
		};
	};
	
} Vector4;

typedef Vector4 Color4;

// Global state
typedef struct
{
	GLint       nWindowWidth;
	GLint		nWindowHeight;
    
    GLint       nRenderTargetWidth;
    GLint       nRenderTargetHeight;
    
    int         nSimulationResolution;
	
	// Pass through
	GLhandleARB hPassThroughVS;
	GLhandleARB hPassThroughFS;
	GLhandleARB hPassThroughShader;
	
	// Draws the gradient into the background texture
	GLhandleARB hGradientFS;
	GLhandleARB hGradientShader;
	
	// Draws the rippls onto the screen
	GLhandleARB hWaterRippleVS;
	GLhandleARB hWaterRippleFS;
	GLhandleARB hWaterRippleShader;
	
	// Injects rain drops into simulation
	GLhandleARB	hRainDropFS;
	GLhandleARB	hRainDropShader;
	
	// Water surface simulation update shaders
	GLhandleARB hSimulationVS;
	GLhandleARB hSimulationFS;
	GLhandleARB hSimulationShader;
    
	// Generates normal map from the simulation
	GLhandleARB hNormalGenFS;
	GLhandleARB hNormalGenShader;
	
	// Global shader params
	GLint		nPassThroughSamplerParamID;
	GLint		nGradientColorShaderParamIDs[2];
	GLint		nBackgroundSamplerID;
	GLint		nTextureResolutionParamID;
	GLint		nTimeStepParamID;
	GLint		nMVPMatrixParamID;
	GLint		nWaterSimSamplerID;
	GLint		nNormalSamplerID;
    GLint       nNormalGenSamplerID;
    GLint       nRippleHeightParamID;
	
	Bool		bResetSim;
	Bool        bPauseSim;
	
	Bool		bRegenerateGradient;
    Bool        bPauseGradient;
    
#if ENABLE_FRAME_DUMPING
    Bool        bDumpMovieFrames;
    Bool        bDumpScreenShot;
#endif // ENABLE_FRAME_DUMPING
	
	// Index of the most recent simulation results
	int			nNewestSimBufferIndex;
	
	// Gradient colors
	Color3		vCoolColors[2];
	Color3		vWarmColors[2];
	Bool        bCoolFirst[2];
    
    // Droplet buffer (age is stored in z)
    Vector3     pDropletList[MAX_DROPLET_COUNT];
	
	// Frame buffer object handles
	GLuint      nNormalFBO;
	GLuint      nWaterSimFBOs[2];
	GLuint      nBackgroundFBO;
    
	// Texture handles
	GLuint		nNormalTexture;
	GLuint		nWaterSimTextures[2];
	GLuint      nBackgroundTexture;
	
	float       fTimeCounter;
	float		fTimeStep;
    
	int         nDebugTextureIndex;
	
	float       mModelViewProjMatrix[16];
	
	GLXContext* pGLXContext;
	
} PuddleStateStruct;

static PuddleStateStruct *g_pPuddleState = NULL;

// ======================================================================
// Vertex shader source
// ======================================================================

// Simple pass through
static const GLchar* g_pStrPassThroughVS = "								\
                                                                            \
    void main()																\
    {																		\
        gl_Position = gl_Vertex;											\
        gl_TexCoord[0].xy = gl_MultiTexCoord0.xy;							\
        gl_TexCoord[0].zw = gl_Vertex.xy * vec2(0.5,0.5) + vec2(0.5,0.5);	\
        gl_TexCoord[1].xy = gl_MultiTexCoord0.zw;							\
    }";

// Simple trasform vertex shader
static const GLchar* g_pStrWaterRippleVS = "								\
                                                                            \
    uniform mat4 g_mMVP; /* ModelViewProj composite */						\
                                                                            \
    void main()																\
    {																		\
        gl_Position = g_mMVP * gl_Vertex;									\
        gl_TexCoord[0].xy = gl_MultiTexCoord0.xy;							\
        gl_TexCoord[0].zw = gl_Position.xy * vec2(0.5,0.5) + vec2(0.5,0.5);	\
    }";

// Water sim vertex shader. Mostly just sets up UVs
static const GLchar* g_pStrWaterSimVS = "                                               \
                                                                                        \
uniform vec2 g_vWaterSimSize;                                                           \
                                                                                        \
    void main()																			\
    {																					\
        gl_Position = gl_Vertex;														\
                                                                                        \
        /* Center */																	\
        gl_TexCoord[0] = gl_MultiTexCoord0;												\
                                                                                        \
        /* Left and Right */															\
        gl_TexCoord[1].xy = gl_MultiTexCoord0.xy + vec2( -1.0/g_vWaterSimSize.x, 0.0 );	\
        gl_TexCoord[1].zw = gl_MultiTexCoord0.xy + vec2(  1.0/g_vWaterSimSize.x, 0.0 );	\
                                                                                        \
        /* Up and Down */																\
        gl_TexCoord[2].xy = gl_MultiTexCoord0.xy + vec2( 0.0,  1.0/g_vWaterSimSize.y );	\
        gl_TexCoord[2].zw = gl_MultiTexCoord0.xy + vec2( 0.0, -1.0/g_vWaterSimSize.y );	\
    }";


// ======================================================================
// Fragment shader source
// ======================================================================

// Simple pass through
static const GLchar* g_pStrPassThroughFS = "									\
                                                                                \
    uniform sampler2D g_sSampler;												\
                                                                                \
    void main()																	\
    {																			\
        gl_FragColor = texture2D( g_sSampler, gl_TexCoord[0].xy  );				\
    }";

// Main ripple screen saver
static const GLchar* g_pStrWaterRippleFS = "									                            \
                                                                                                            \
    uniform sampler2D g_sBackgroundSampler;										                            \
    uniform sampler2D g_sNormalSampler;											                            \
                                                                                                            \
    void main()																	                            \
    {                                                                                                       \
                                                                                                            \
        vec3 vNormal = texture2D( g_sNormalSampler, gl_TexCoord[0].xy ).xyz * vec3(2,2,2) - vec3(1,1,1);	\
        vNormal = normalize( vNormal );																		\
                                                                                                            \
        /* Mask ripples out at the top of the screen.  The creates the illusion of the puddle quad sort */	\
        /* of fading or fogging out at the top of the screen and also helps with some aliasing          */	\
        float fMask = 1.0 - ( smoothstep(0.65, 1.0, clamp(gl_TexCoord[0].w,0.0,1.0)) * 0.5 );				\
        vec2  vMask = vec2( fMask, fMask );																	\
        vec2  vInvMask = vec2( 1, 1 ) - vMask;																\
                                                                                                            \
        vec2 vSharpenExp = vec2(30,30) * vInvMask + vec2(8,8);												\
        vec2 vUV = gl_TexCoord[0].zw + clamp( vNormal.xy * pow(vNormal.zz,vSharpenExp), -vMask, vMask );	\
        vUV = clamp( vUV, vec2(0,0), vec2(1,1) );															\
        gl_FragColor = texture2D( g_sBackgroundSampler, vUV );												\
    }";

// Two color non-linear gradient
static const GLchar* g_pStrTwoColorGradientFS = "							\
                                                                            \
    uniform vec4 g_vColor0;													\
    uniform vec4 g_vColor1;													\
                                                                            \
    void main()																\
    {																		\
        float fWeight = smoothstep(0.0, 1.0, gl_TexCoord[0].y );			\
        gl_FragColor = mix( g_vColor0, g_vColor1, fWeight );				\
    }";

// Injects some energy into the sim
static const GLchar* g_pStrRainDropFS = "                                                   \
                                                                                            \
    void main()																				\
    {                                                                                       \
        float f3PI = 3.14159 * 3.0;                                                         \
        float fDropletAge = gl_TexCoord[1].y;                                               \
                                                                                            \
        /* Normalized distance from center of the rain drop quad [0,1] */					\
        float fDist = clamp( length(gl_TexCoord[0].xy * vec2(2,2) - vec2(1,1)), 0.0, 1.0 ); \
        float fMask = clamp(1.0 - fDist*1.1 - 0.1, 0.0, 1.0) ;                              \
                                                                                            \
        float fPulse = sin( fDropletAge * f3PI ) * fMask * 2.0 + 0.5;                       \
        float fEncodedPulse = 0.5 + 0.5 * clamp( fPulse, -1.0, 1.0 );						\
                                                                                            \
        gl_FragColor = vec4(fEncodedPulse,fEncodedPulse,fEncodedPulse,fMask);               \
    }";



// Water sim fragment shader
static const GLchar* g_pStrWaterSimFS = "                                                                       \
                                                                                                                \
    uniform float g_fTimeStep;																					\
    uniform sampler2D g_sWaterSimSampler;																		\
                                                                                                                \
    void main()																									\
    {																											\
        const float fCoeffSquared = 200.0;																		\
        const float fSpring = 0.05;																				\
        const float fDamping = 0.995;																			\
        float fDeltaTime = g_fTimeStep;																			\
                                                                                                                \
        vec3 vNewElement;																						\
        vec3 vElement;																							\
                                                                                                                \
        /* Grab the value of the currently simulated vElement */												\
        vElement = texture2D( g_sWaterSimSampler, gl_TexCoord[0].xy ).xyz;										\
                                                                                                                \
                                                                                                                \
        /* Compute the acceleration factor	*/																	\
        vNewElement.z = fCoeffSquared * ( texture2D( g_sWaterSimSampler, gl_TexCoord[1].xy ).x +  /* Left	*/	\
        texture2D( g_sWaterSimSampler, gl_TexCoord[1].zw ).x +  /* Right  */                                    \
        texture2D( g_sWaterSimSampler, gl_TexCoord[2].xy ).x +  /* Top	*/                                      \
        texture2D( g_sWaterSimSampler, gl_TexCoord[2].zw ).x -  /* Bottom */                                    \
        4.0 * vElement.x ) +                                                                                    \
        fSpring * (0.5 - vElement.x );  /* low-force spring to pull water surface back to rest position */      \
                                                                                                                \
        /* Integrate to get velocity */																			\
        vNewElement.y = ( vElement.y - 0.5 ) + fDeltaTime * vNewElement.z;										\
                                                                                                                \
        /* Integrate again to get position */																	\
        vNewElement.x = ( vElement.x - 0.5 ) + fDeltaTime * vNewElement.y;										\
                                                                                                                \
        /* Apply damping to let the system lose some energy	*/													\
        vNewElement.y *= fDamping;																				\
                                                                                                                \
        /* Bias	*/																								\
        gl_FragColor.rgb = vNewElement + vec3(0.5, 0.5, 0.5);													\
        gl_FragColor.a = 1.0;																					\
    }";

// Forward differencing to derive per-pixel normal
static const GLchar* g_pStrNormalGenFS = "                                                                                      \
                                                                                                                                \
    uniform sampler2D g_sWaterSimSampler;                                                                                       \
    uniform vec2 g_vRippleHeight; /* <height, 1/height>  */                                                                     \
                                                                                                                                \
    void main()                                                                                                                 \
    {                                                                                                                           \
        vec3 vNormal = vec3( 0, 0, g_vRippleHeight.y );                                                                         \
        vec3 vCenter = texture2D( g_sWaterSimSampler, gl_TexCoord[0].xy ).xyz; /* center tap */                                 \
                                                                                                                                \
        vNormal.x = -g_vRippleHeight.x * (vCenter.x - texture2D( g_sWaterSimSampler, gl_TexCoord[1].xy ).x ); /* left tap */	\
        vNormal.y = -g_vRippleHeight.x * (vCenter.x - texture2D( g_sWaterSimSampler, gl_TexCoord[2].xy ).x ); /* top tap */     \
                                                                                                                                \
        /* scale and bias to [0.1] for storing in unsigned texture */                                                           \
        gl_FragColor.rgb = normalize( vNormal ) * vec3(0.5, 0.5, 0.5) + vec3(0.5, 0.5, 0.5);                                    \
        gl_FragColor.a = 1.0;                                                                                                   \
    }";



// ======================================================================
// Test for errors
// ======================================================================
static Bool PrintShaderCompilationLog ( const char* pStrShaderName, GLhandleARB hShaderObject )
{
	GLchar strMessage[2048];
	GLsizei nMessageLength = 0;
	
    glGetInfoLogARB( hShaderObject, (sizeof strMessage) / (sizeof strMessage[0]), &nMessageLength, strMessage );
    
	if ( 0 < nMessageLength )
	{
		fprintf( stderr, "%s :\n %s", (NULL==pStrShaderName) ? "Unknown" : pStrShaderName, strMessage );
		return FALSE;
	}
	
	return TRUE;
}

// ==================================================================================
// OpenGL debugging is a pain in the ass. Why can't we just have a debug runtime?
// WARNING: Do NOT wrap glBegin/glEnd (or anything inbetween those calls) with
//          GL_DEBUG macros, it is illegal to check for GL errors inside a
//          glBegin/glEnd block.
// ==================================================================================
#if ENABLE_GL_DEBUGGING

static Bool PrintGLErrorVerbose ( const char* pStrLabel, const char* pStrFileName, const int nLineNumber )
{
    GLenum eErrorCode = GL_NO_ERROR;
    if ( ( eErrorCode = glGetError() ) != GL_NO_ERROR )
    {
        const GLubyte *pStrError = gluErrorString(eErrorCode);
        
        if ( NULL != pStrFileName )
        {
            fprintf( stderr, "OpenGL Error: \n\tCause: %s\n\tResult: %s (0x%x)\n\tLocation: %s (%d)\n",
                    pStrLabel ? pStrLabel : "Unknown", pStrError ? (const char*)pStrError : "Unknown", eErrorCode, pStrFileName, nLineNumber );
        }
        else
        {
            fprintf( stderr, "OpenGL Error: \n\tCause: %s\n\tResult: %s (0x%x)\n", pStrLabel ? pStrLabel : "Unknown", pStrError ? (const char*)pStrError : "Unknown", eErrorCode );
        }
        
        return FALSE;
    }
    
    return TRUE;
}

// Simple version
static Bool PrintGLError ( const char* pStrLabel )
{
    return PrintGLErrorVerbose( pStrLabel, NULL, 0 );
}

#define GL_DEBUG(expression) { expression; PrintGLErrorVerbose(#expression, __FILE__, __LINE__); }
#define GL_DEBUG_LABEL(label) { PrintGLErrorVerbose(#label, __FILE__, __LINE__); }
#define DEBUG_LABEL(label) { fprintf( stderr, "Debug Label: %s\tLocation: %s (%d)\n", #label, __FILE__, __LINE__ ); }
#else // !ENABLE_GL_DEBUGGING

#define GL_DEBUG(expression) expression
#define GL_DEBUG_LABEL(label)
#define DEBUG_LABEL(lable)
#endif // ENABLE_GL_DEBUGGING

//=======================================================
// Linear interpolation.  fWeight is assumed to be [0,1]
//=======================================================
static inline float Lerp ( float fA, float fB, float fWeight )
{
	return ((1.0f-fWeight) * fA) + (fWeight * fB);
}


// ======================================================================
// Linear interpolation between two colors. It is safe to overlap input
// and output colors.
// ======================================================================
static void LerpColor3 ( Color3* pOutput, Color3* pColorA, Color3* pColorB, float fWeight )
{
	pOutput->r = Lerp( pColorA->r, pColorB->r, fWeight );
	pOutput->g = Lerp( pColorA->g, pColorB->g, fWeight );
	pOutput->b = Lerp( pColorA->b, pColorB->b, fWeight );
}

//=====================
// Cubic interpolation
//=====================
static float SmoothStep ( float fMin, float fMax, float fX )
{
	if ( fX < fMin )
	{
		return 0.f;
	}
	
	if ( fX >= fMax )
	{
		return 1.f;
	}
	
	fX = ( fX - fMin) / ( fMax - fMin );
	return ( fX * fX * ( 3.f - 2.f * fX ) );
}

//===============================
// Uniform random number in [0,1]
//===============================
inline static float UniformRandomNumber()
{
    return LRAND() / (float)MAXRAND;
}

//=========================================================================
// Poisson random number. Lambda is expected result over unit time interval
//=========================================================================
inline static int PoissonRandomNumber ( float fLambda )
{
    // Cribbed from: http://en.wikipedia.org/wiki/Poisson_distribution
    
    int nK = 0;
    const float fTarget = exp( -fLambda );
    float fP = UniformRandomNumber();
    
    while ( fP > fTarget)
    {
        fP *= UniformRandomNumber();
        nK += 1;
    }
    
    return nK;
}

//====================================================================================================================
// Computes Cool and Warm color tones for Gooch style shading. Takes a reference color (and other params) as input and
// generated an output cool and warm color pair.
//====================================================================================================================
static void GenerateGoochTones ( float fBlue, float fYellow, float fColorStrength, float fLuminanceShift, const Color3* pRefColor, Color3* pCoolColorOutput, Color3* pWarmColorOutput )
{
	pCoolColorOutput->r = fColorStrength * pRefColor->r;
	pCoolColorOutput->g = fColorStrength * pRefColor->g;
	pCoolColorOutput->b = fBlue + (fColorStrength * pRefColor->b);
	
	pWarmColorOutput->r = fYellow + (fLuminanceShift * pRefColor->r);
	pWarmColorOutput->g = fYellow + (fLuminanceShift * pRefColor->g);
	pWarmColorOutput->b = fLuminanceShift * pRefColor->b;
}

//======================================================================================
// Computes a warm and cool color at random. See above for explanation of inputs/outputs.
//======================================================================================
static void GenerateRandomGoochTones ( Color3* pCool, Color3* pWarm )
{
	Color3 color;
	color.r = UniformRandomNumber();
	color.g = UniformRandomNumber();
	color.b = UniformRandomNumber();
	
	GenerateGoochTones( UniformRandomNumber(), UniformRandomNumber(), UniformRandomNumber(), UniformRandomNumber(), &color, pCool, pWarm );
}

// ======================================================================
// TopX, TopY, Width, and Height are all in NDC [0,1] space
// ======================================================================
static void VisualizeTexture ( PuddleStateStruct* pState, GLint nTextureID, float fTopLeftX, float fTopLeftY, float fWidth, float fHeight  )
{
	GL_DEBUG( glDisable( GL_DEPTH_TEST ) );
	GL_DEBUG( glDisable( GL_CULL_FACE ) );
	
	// Bind the pass through shader
	GL_DEBUG( glUseProgramObjectARB( pState->hPassThroughShader ) );
    
	// Hook up the sampler
	GL_DEBUG( glActiveTexture( GL_TEXTURE0 + 0 ) );
	GL_DEBUG( glBindTexture( GL_TEXTURE_2D, nTextureID ) );
	GL_DEBUG( glUniform1iARB( pState->nPassThroughSamplerParamID, 0 ) );
	
	// Calculate screen space coords
	float fSSTopLeftX     = fTopLeftX * 2.0f - 1.0f;
	float fSSTopLeftY     = fTopLeftY * 2.0f - 1.0f;
	
	float fSSBottomRightX = fSSTopLeftX + (fWidth * 2.0f);
	float fSSBottomRightY = fSSTopLeftY - (fHeight * 2.0f);
    
    glBegin(GL_QUADS);
	
    glTexCoord4f( 0.f, 0.f, 0.f, 1.f );
    glVertex4f( fSSTopLeftX, fSSBottomRightY, 0.5f, 1 );
    
    glTexCoord4f( 1.f, 0.f, 0.f, 1.f );
    glVertex4f( fSSBottomRightX, fSSBottomRightY, 0.5f, 1.f );
    
    glTexCoord4f( 1.f, 1.f, 0.f, 1.f );
    glVertex4f( fSSBottomRightX, fSSTopLeftY, 0.5f, 1.f );
    
    glTexCoord4f( 0.f, 1.f, 0.f, 1.f );
    glVertex4f( fSSTopLeftX, fSSTopLeftY, 0.5f, 1.f );
	
	glEnd();
    
    // Unbind the texture and turn the programmable pipeline off
	GL_DEBUG( glBindTexture( GL_TEXTURE_2D, 0 ) );
	GL_DEBUG( glUseProgramObjectARB( 0 ) );
}

// ======================================================================
// Debug visualization
// ======================================================================
static void DrawDebugVisualization ( PuddleStateStruct* pState )
{
	// Debug render target
	switch ( pState->nDebugTextureIndex )
	{
		case 0 :
            VisualizeTexture( pState, pState->nNormalTexture, 0.125f, 0.875f, 0.75f, 0.75f );
			break;
			
		case 1 :
            VisualizeTexture( pState, pState->nWaterSimTextures[0], 0.125f, 0.875f, 0.75f, 0.75f );
			break;
			
		case 2 :
            VisualizeTexture( pState, pState->nWaterSimTextures[1], 0.125f, 0.875f, 0.75f, 0.75f );
			break;
			
		case 3 :
            VisualizeTexture( pState, pState->nBackgroundTexture, 0.125f, 0.875f, 0.75f, 0.75f );
			break;
			
		case 4 :
            VisualizeTexture( pState, pState->nNormalTexture,       0.0625f,   0.9375f,   0.421875f, 0.421875f );
            VisualizeTexture( pState, pState->nWaterSimTextures[0], 0.515625f, 0.9375f,   0.421875f, 0.421875f );
            VisualizeTexture( pState, pState->nWaterSimTextures[1], 0.515625f, 0.484375f, 0.421875f, 0.421875f );
            VisualizeTexture( pState, pState->nBackgroundTexture,   0.0625f,   0.484375f, 0.421875f, 0.421875f );
            break;
			
		default :
            pState->nDebugTextureIndex = -1; // Wrap back to an 'off' state
			break;
	}
}

// ======================================================================
//
// ======================================================================
static void Init ( ModeInfo *pMI )
{
	PuddleStateStruct *pState = &g_pPuddleState[MI_SCREEN(pMI)];
    
    glDisable( GL_LIGHTING );
	glDisable( GL_DEPTH_TEST );
	glDisable( GL_CULL_FACE );
    
    glClearColor(0.f, 0.f, 0.f, 1.f);
	
	pState->fTimeCounter = 0.f;
	pState->fTimeStep = 0.f;
	pState->bResetSim = TRUE; // Simulation will be reset on the first frame
	pState->bPauseSim = FALSE;
	pState->bRegenerateGradient = TRUE;
    pState->bPauseGradient = FALSE;
	pState->nNewestSimBufferIndex = 0;
	pState->nDebugTextureIndex = -1;
    
#if ENABLE_FRAME_DUMPING
    pState->bDumpMovieFrames = FALSE;
    pState->bDumpScreenShot = FALSE;
#endif // ENABLE_FRAME_DUMPING
    
    pState->nSimulationResolution = (256 << g_nSimResolutionEnum); //0=256, 1=512, 2=1024, 3=2048
    
    // Rember the resolution we created the render targets at since we won't be
    // recreating them if the window size changes
    pState->nRenderTargetWidth = pState->nWindowWidth;
    pState->nRenderTargetHeight = pState->nWindowHeight;
	
	// Setup colors for gradient
	GenerateRandomGoochTones( &(pState->vWarmColors[0]), &(pState->vCoolColors[0]) );
	GenerateRandomGoochTones( &(pState->vWarmColors[1]), &(pState->vCoolColors[1]) );
	pState->bCoolFirst[0] = pState->bCoolFirst[1] = TRUE;
    
	// Create textures
	GL_DEBUG( glGenTextures( 1, &pState->nNormalTexture ) );
	GL_DEBUG( glGenTextures( 2, &pState->nWaterSimTextures[0] ) );
	GL_DEBUG( glGenTextures( 1, &pState->nBackgroundTexture ) );
	
	GLfloat fLargestAnisoKernelSize = 0.f;
	
	// Water sim's normal map
	GL_DEBUG( glBindTexture( GL_TEXTURE_2D, pState->nNormalTexture ) );
	GL_DEBUG( glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR ) );
	GL_DEBUG( glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR ) );
	GL_DEBUG( glGetFloatv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &fLargestAnisoKernelSize) );
	GL_DEBUG( glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, fLargestAnisoKernelSize) );
	GL_DEBUG( glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT ) );
	GL_DEBUG( glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT ) );
	//GL_DEBUG( glTexParameteri( GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE ) ); // Automatic mipmap
	GL_DEBUG( glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA8, pState->nSimulationResolution, pState->nSimulationResolution, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0) );
    //GL_DEBUG( glTexStorage2D( GL_TEXTURE_2D, log2(pState->nSimulationResolution), GL_RGBA8, pState->nSimulationResolution, pState->nSimulationResolution ) );
    GL_DEBUG( glGenerateMipmap(GL_TEXTURE_2D) );
    GL_DEBUG( glBindTexture(GL_TEXTURE_2D, 0) );
	
	// TODO: Pick a sim buffer format at runtime.
    // Not sure which is more likely to be supported, fp16 or uint16? We need alpha blending to be supported for
    // whatever format we go with. How do you test for this at runtime in OpenGL?!
    // The 10 & 8 bit formats will exhibit noise from precision error and generally fall apart, but I guess they are
    // better than nothing.
    
    int nSimTexInternalFormat, nSimTexType;
    
    switch ( SIMULATION_TEXTURE_FORMAT )
    {
        case 0 :
            nSimTexInternalFormat = GL_RGBA8;
            nSimTexType = GL_UNSIGNED_BYTE;
            break;
            
        case 1 :
            nSimTexInternalFormat = GL_RGB10_A2;
            nSimTexType = GL_UNSIGNED_INT_10_10_10_2;
            break;
            
        case 2 :
            nSimTexInternalFormat = GL_RGBA16;
            nSimTexType = GL_UNSIGNED_SHORT;
            break;
            
        case 3 :
            nSimTexInternalFormat = GL_RGBA16F_ARB;
            nSimTexType = GL_FLOAT;
            break;
            
        case 4 :
            nSimTexInternalFormat = GL_RGBA32F_ARB;
            nSimTexType = GL_FLOAT;
            break;
            
        default:
            fprintf( stderr, "Invalid SIMULATION_TEXTURE_FORMAT (%d)! I'll just pick a default for you!\n", SIMULATION_TEXTURE_FORMAT);
            nSimTexInternalFormat = GL_RGBA16;
            nSimTexType = GL_UNSIGNED_SHORT;
            break;
    }
	
	int nBuffer=0;
	for ( nBuffer=0; nBuffer<2; nBuffer++ )
	{
		// Water sim buffers
		GL_DEBUG( glBindTexture( GL_TEXTURE_2D, pState->nWaterSimTextures[nBuffer] ) );
		GL_DEBUG( glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST ) );
		GL_DEBUG( glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST ) );
		GL_DEBUG( glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT ) );
		GL_DEBUG( glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT ) );
		GL_DEBUG( glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL, 0) );
		GL_DEBUG( glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, 0) );
		GL_DEBUG( glTexParameteri( GL_TEXTURE_2D, GL_GENERATE_MIPMAP, FALSE ) );
		GL_DEBUG( glTexImage2D( GL_TEXTURE_2D, 0, nSimTexInternalFormat, pState->nSimulationResolution, pState->nSimulationResolution, 0, GL_RGBA, nSimTexType, 0) );
		GL_DEBUG( glBindTexture(GL_TEXTURE_2D, 0) );	}
	
	// Background image
	GL_DEBUG( glBindTexture( GL_TEXTURE_2D, pState->nBackgroundTexture ) );
	GL_DEBUG( glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR ) );
	GL_DEBUG( glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, ENABLE_BACKGROUND_MIP ? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR ) );
	GL_DEBUG( glGetFloatv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &fLargestAnisoKernelSize) );
	GL_DEBUG( glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, fLargestAnisoKernelSize) );
	GL_DEBUG( glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE ) );
	GL_DEBUG( glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE ) );
	//GL_DEBUG( glTexParameteri( GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE ) ); // Automatic mipmap
	GL_DEBUG( glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA8, 1, pState->nRenderTargetHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0) );
    //GL_DEBUG( glTexStorage2D( GL_TEXTURE_2D, (GLsizei)log2(pState->nRenderTargetHeight), GL_RGBA8, pState->nSimulationResolution, pState->nSimulationResolution ) );
    if ( ENABLE_BACKGROUND_MIP )
    {
        GL_DEBUG( glGenerateMipmap(GL_TEXTURE_2D) );
    }
	GL_DEBUG( glBindTexture(GL_TEXTURE_2D, 0) );
	
	// Create frame buffer objects
	GL_DEBUG( glGenFramebuffersEXT( 1, &pState->nNormalFBO ) );
	GL_DEBUG( glGenFramebuffersEXT( 2, &pState->nWaterSimFBOs[0] ) );
	GL_DEBUG( glGenFramebuffersEXT( 1, &pState->nBackgroundFBO ) );
	
	// Attach the normal texture to its FBO
	GL_DEBUG( glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, pState->nNormalFBO ) );
	GL_DEBUG( glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, pState->nNormalTexture, 0 ) );
    
	// Check FBO status
	GLenum eFBOStatus = glCheckFramebufferStatusEXT( GL_FRAMEBUFFER_EXT );
	if ( eFBOStatus != GL_FRAMEBUFFER_COMPLETE_EXT )
	{
		fprintf( stderr, "Failed to create FBO for water sim normal texture (return code = 0x%x)\n", eFBOStatus );
	}
	
	GL_DEBUG( glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, 0 ) );
    
	
	// Attach the water sim textures to their FBOs
	for ( nBuffer=0; nBuffer<2; nBuffer++ )
	{
		GL_DEBUG( glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, pState->nWaterSimFBOs[nBuffer] ) );
		GL_DEBUG( glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, pState->nWaterSimTextures[nBuffer], 0 ) );
        
		// Check FBO status
		eFBOStatus = glCheckFramebufferStatusEXT( GL_FRAMEBUFFER_EXT );
		if ( eFBOStatus != GL_FRAMEBUFFER_COMPLETE_EXT )
		{
			fprintf( stderr, "Failed to create FBO for water sim texture (return code = 0x%x)\n", eFBOStatus );
		}
        
		GL_DEBUG( glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0) );
	}
	
	// Attach the background texture to its FBO
	GL_DEBUG( glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, pState->nBackgroundFBO ) );
	GL_DEBUG( glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, pState->nBackgroundTexture, 0 ) );
	
	// Check FBO status
	eFBOStatus = glCheckFramebufferStatusEXT( GL_FRAMEBUFFER_EXT );
	if ( eFBOStatus != GL_FRAMEBUFFER_COMPLETE_EXT )
	{
		fprintf( stderr, "Failed to create FBO for background texture\n" );
	}
	
	GL_DEBUG( glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, 0 ) );
    
	// Create Shader And Program Objects
	pState->hPassThroughShader = glCreateProgramObjectARB();
	pState->hGradientShader = glCreateProgramObjectARB();
	pState->hWaterRippleShader = glCreateProgramObjectARB();
	pState->hSimulationShader = glCreateProgramObjectARB();
	pState->hRainDropShader = glCreateProgramObjectARB();
	pState->hNormalGenShader = glCreateProgramObjectARB();
	GL_DEBUG_LABEL( "Create Program Objects" );
	
	pState->hPassThroughVS = glCreateShaderObjectARB( GL_VERTEX_SHADER_ARB );
	pState->hPassThroughFS = glCreateShaderObjectARB( GL_FRAGMENT_SHADER_ARB );
	pState->hGradientFS = glCreateShaderObjectARB( GL_FRAGMENT_SHADER_ARB );
	pState->hWaterRippleFS = glCreateShaderObjectARB( GL_FRAGMENT_SHADER_ARB );
	pState->hWaterRippleVS = glCreateShaderObjectARB( GL_VERTEX_SHADER_ARB );
	pState->hSimulationVS = glCreateShaderObjectARB( GL_VERTEX_SHADER_ARB );
	pState->hSimulationFS = glCreateShaderObjectARB( GL_FRAGMENT_SHADER_ARB );
	pState->hNormalGenFS = glCreateShaderObjectARB( GL_FRAGMENT_SHADER_ARB );
	pState->hRainDropFS = glCreateShaderObjectARB( GL_FRAGMENT_SHADER_ARB );
	GL_DEBUG_LABEL( "Create Shader Objects" );
	
	// Load Shader Sources
	GL_DEBUG( glShaderSourceARB( pState->hPassThroughVS, 1, &g_pStrPassThroughVS, NULL ) );
	GL_DEBUG( glShaderSourceARB( pState->hPassThroughFS, 1, &g_pStrPassThroughFS, NULL ) );
	GL_DEBUG( glShaderSourceARB( pState->hGradientFS, 1, &g_pStrTwoColorGradientFS, NULL ) );
	GL_DEBUG( glShaderSourceARB( pState->hWaterRippleFS, 1, &g_pStrWaterRippleFS, NULL ) );
	GL_DEBUG( glShaderSourceARB( pState->hWaterRippleVS, 1, &g_pStrWaterRippleVS, NULL ) );
	GL_DEBUG( glShaderSourceARB( pState->hSimulationVS, 1, &g_pStrWaterSimVS, NULL ) );
	GL_DEBUG( glShaderSourceARB( pState->hSimulationFS, 1, &g_pStrWaterSimFS, NULL ) );
	GL_DEBUG( glShaderSourceARB( pState->hRainDropFS, 1, &g_pStrRainDropFS, NULL ) );
	GL_DEBUG( glShaderSourceARB( pState->hNormalGenFS, 1, &g_pStrNormalGenFS, NULL ) );
    
	// Compile the shaders
	GL_DEBUG( glCompileShaderARB( pState->hPassThroughVS ) );
	PrintShaderCompilationLog( "PassThrough VS", pState->hPassThroughVS );
	
	GL_DEBUG( glCompileShaderARB( pState->hPassThroughFS ) );
	PrintShaderCompilationLog( "PassThrough FS", pState->hPassThroughFS );
	
	GL_DEBUG( glCompileShaderARB( pState->hGradientFS ) );
	PrintShaderCompilationLog( "Gradient FS", pState->hGradientFS );
	
	GL_DEBUG( glCompileShaderARB( pState->hWaterRippleVS ) );
	PrintShaderCompilationLog( "WaterRipple VS", pState->hWaterRippleVS );
	
	GL_DEBUG( glCompileShaderARB( pState->hWaterRippleFS ) );
	PrintShaderCompilationLog( "WaterRipple FS", pState->hWaterRippleFS );
	
	GL_DEBUG( glCompileShaderARB( pState->hSimulationVS ) );
	PrintShaderCompilationLog( "Simulation VS", pState->hSimulationVS );
	
	GL_DEBUG( glCompileShaderARB( pState->hSimulationFS ) );
	PrintShaderCompilationLog( "Simulation FS", pState->hSimulationFS );
	
	GL_DEBUG( glCompileShaderARB( pState->hRainDropFS ) );
	PrintShaderCompilationLog( "Rain Drop FS", pState->hRainDropFS );
	
	GL_DEBUG( glCompileShaderARB( pState->hNormalGenFS ) );
	PrintShaderCompilationLog( "Normal Generation FS", pState->hNormalGenFS );
	
	// Attach the shader objects to the program object
	GL_DEBUG( glAttachObjectARB( pState->hPassThroughShader, pState->hPassThroughVS ) );
	GL_DEBUG( glAttachObjectARB( pState->hPassThroughShader, pState->hPassThroughFS ) );
	
	GL_DEBUG( glAttachObjectARB( pState->hGradientShader, pState->hPassThroughVS ) );
	GL_DEBUG( glAttachObjectARB( pState->hGradientShader, pState->hGradientFS ) );
	
	GL_DEBUG( glAttachObjectARB( pState->hWaterRippleShader, pState->hWaterRippleVS ) );
	GL_DEBUG( glAttachObjectARB( pState->hWaterRippleShader, pState->hWaterRippleFS ) );
	
	GL_DEBUG( glAttachObjectARB( pState->hSimulationShader, pState->hSimulationVS ) );
	GL_DEBUG( glAttachObjectARB( pState->hSimulationShader, pState->hSimulationFS ) );
	
	GL_DEBUG( glAttachObjectARB( pState->hRainDropShader, pState->hPassThroughVS ) );
	GL_DEBUG( glAttachObjectARB( pState->hRainDropShader, pState->hRainDropFS ) );
	
	GL_DEBUG( glAttachObjectARB( pState->hNormalGenShader, pState->hSimulationVS ) );
	GL_DEBUG( glAttachObjectARB( pState->hNormalGenShader, pState->hNormalGenFS ) );
    
	// Link the program objects
	GL_DEBUG( glLinkProgramARB( pState->hPassThroughShader ) );
	PrintShaderCompilationLog( "Pass Through Shader", pState->hPassThroughShader );
	
	GL_DEBUG( glLinkProgramARB( pState->hGradientShader ) );
	PrintShaderCompilationLog( "Gradient Shader", pState->hGradientShader );
	
	GL_DEBUG( glLinkProgramARB( pState->hWaterRippleShader ) );
	PrintShaderCompilationLog( "Water Ripple Shader", pState->hWaterRippleShader );
	
	GL_DEBUG( glLinkProgramARB( pState->hSimulationShader ) );
	PrintShaderCompilationLog( "Simulation Shader", pState->hSimulationShader );
	
	GL_DEBUG( glLinkProgramARB( pState->hRainDropShader ) );
	PrintShaderCompilationLog( "Rain Drop Shader", pState->hRainDropShader );
	
	GL_DEBUG( glLinkProgramARB( pState->hNormalGenShader ) );
	PrintShaderCompilationLog( "Normal Generation Shader", pState->hNormalGenShader );
	
	// Get the pass through sampler param
	pState->nPassThroughSamplerParamID = glGetUniformLocationARB( pState->hPassThroughShader, "g_sSampler" );
	GL_DEBUG_LABEL( "Get the pass through sampler param" );
	
	// Get the colors for the gradient params
	pState->nGradientColorShaderParamIDs[0] = glGetUniformLocationARB( pState->hGradientShader, "g_vColor0" );
	pState->nGradientColorShaderParamIDs[1] = glGetUniformLocationARB( pState->hGradientShader, "g_vColor1" );
	GL_DEBUG_LABEL( "Get the colors for the gradient params" );
	
	// Get the background sampler param
	pState->nBackgroundSamplerID = glGetUniformLocationARB( pState->hWaterRippleShader, "g_sBackgroundSampler" );
	GL_DEBUG_LABEL( "Get the background sampler param" );
	
	// Get the background sampler param
	pState->nNormalSamplerID = glGetUniformLocationARB( pState->hWaterRippleShader, "g_sNormalSampler" );
	pState->nMVPMatrixParamID = glGetUniformLocationARB( pState->hWaterRippleShader, "g_mMVP" );
	GL_DEBUG_LABEL( "Get the normal sampler param" );
    
	// Get simulation params
	pState->nTextureResolutionParamID = glGetUniformLocationARB( pState->hSimulationShader, "g_vWaterSimSize" );
	pState->nTimeStepParamID = glGetUniformLocationARB( pState->hSimulationShader, "g_fTimeStep" );
	pState->nWaterSimSamplerID = glGetUniformLocationARB( pState->hSimulationShader, "g_sWaterSimSampler" );
	GL_DEBUG_LABEL( "Get the simulation params" );
    
    // Get the normal gen params
	pState->nNormalGenSamplerID = glGetUniformLocationARB( pState->hNormalGenShader, "g_sWaterSimSampler" );
    pState->nRippleHeightParamID = glGetUniformLocationARB( pState->hNormalGenShader, "g_vRippleHeight" );
	GL_DEBUG_LABEL( "Get the normal gen params" );
	
}

// ======================================================================
// Helper function to draw screen aligned quads
// ======================================================================
static void DrawScreenQuad ( PuddleStateStruct* pState, Vector3* pScale, Vector3* pBias, Vector4* pUV )
{
	Vector3 vScale;
	Vector3 vBias;
	Vector4 vUV;
    
	if ( pScale )
	{
		vScale = *pScale;
	}
	else
	{
		vScale.x = vScale.y = vScale.z = 1.f;
		
	}
    
	if ( pBias )
	{
		vBias = *pBias;
	}
	else
	{
		vBias.x = vBias.y = vBias.z = 0.f;
	}
	
	if ( pUV )
	{
		vUV = *pUV;
	}
	else
	{
		vUV.x = vUV.y = 0.f;
		vUV.z = vUV.w = 1.f;
	}
    
    glBegin(GL_QUADS);
    
    glTexCoord4f(vUV.x, vUV.y, 0.f, 1.f);
    glVertex4f(-1.f * vScale.x + vBias.x, -1.f * vScale.y + vBias.y, 0.5f * vScale.z + vBias.z, 1);
    
    glTexCoord4f(vUV.z, vUV.y, 0.f, 1.f);
    glVertex4f( 1.f * vScale.x + vBias.x, -1.f * vScale.y + vBias.y, 0.5f * vScale.z + vBias.z, 1);
    
    glTexCoord4f(vUV.z, vUV.w, 0.f, 1.f);
    glVertex4f( 1.f * vScale.x + vBias.x, 1.f * vScale.y + vBias.y, 0.5f * vScale.z + vBias.z, 1);
    
    glTexCoord4f(vUV.x, vUV.w, 0.f, 1.f);
    glVertex4f(-1.f * vScale.x + vBias.x, 1.f * vScale.y + vBias.y, 0.5f * vScale.z + vBias.z, 1);
    
	glEnd();
}

// ======================================================================
// Draw the ground plane geometry with the ripple shader
// ======================================================================
static void DrawRipplePlane ( PuddleStateStruct* pState )
{
	// Bind the water ripple shader
	GL_DEBUG( glUseProgramObjectARB( pState->hWaterRippleShader ) );
	
	// Set the MVP transform
	GL_DEBUG( glUniformMatrix4fv( pState->nMVPMatrixParamID, 1, GL_FALSE, pState->mModelViewProjMatrix ) );
	
	// Hook up the background and normal textures and trigger their MIP generation
	GL_DEBUG( glActiveTexture( GL_TEXTURE0 + 0 ) );
	GL_DEBUG( glBindTexture( GL_TEXTURE_2D, pState->nBackgroundTexture ) );
    if ( ENABLE_BACKGROUND_MIP )
    {
        GL_DEBUG( glGenerateMipmapEXT( GL_TEXTURE_2D ) );
    }
	GL_DEBUG( glUniform1iARB( pState->nBackgroundSamplerID, 0 ) );
	
	GL_DEBUG( glActiveTexture( GL_TEXTURE0 + 1 ) );
	GL_DEBUG( glBindTexture( GL_TEXTURE_2D, pState->nNormalTexture ) );
	GL_DEBUG( glGenerateMipmapEXT( GL_TEXTURE_2D ) );
	GL_DEBUG( glUniform1iARB( pState->nNormalSamplerID, 1 ) );
	
	// Repeat more frequently for smaller size buffers so that things don't go all pixelated as the sim resolution is decreased
	float fRepeatRate = (16.f*1024.f)/(float)pState->nSimulationResolution;
	
    glBegin(GL_QUADS);
	
    glTexCoord4f(0.f, 0.f, 0.f, 1.f);
    glVertex4f(-4.0f, -4.0f, 0.0f, 1.0f);
    
    glTexCoord4f(fRepeatRate, 0.f, 0.f, 1.f);
    glVertex4f(4.0f, -4.0f, 0.0f, 1.0f);
    
    glTexCoord4f(fRepeatRate, fRepeatRate, 0.f, 1.f);
    glVertex4f(4.0f, 4.0f, 0.0f, 1.0f);
    
    glTexCoord4f(0.f, fRepeatRate, 0.f, 1.f);
    glVertex4f(-4.0f, 4.0f, 0.0f, 1.0f );
	
	glEnd();
    
    
}


// ======================================================================
// Makes a warm/cool color gradient using Gooch tones
// ======================================================================
static void UpdateColorGradient ( PuddleStateStruct* pState )
{
	// Every so many seconds, transition to a new cool/warm color gradient
	const float fTransitionTime = g_fGradientTransitionTime;
	pState->bRegenerateGradient = (pState->fTimeCounter > fTransitionTime) ? TRUE : pState->bRegenerateGradient;
	
	if ( pState->bRegenerateGradient )
	{
		pState->fTimeCounter = 0.0f;
		pState->bRegenerateGradient = FALSE;
		
		pState->vCoolColors[0] = pState->vCoolColors[1];
		pState->vWarmColors[0] = pState->vWarmColors[1];
		pState->bCoolFirst[0] = pState->bCoolFirst[1];
		
		GenerateRandomGoochTones( &pState->vCoolColors[1], &pState->vWarmColors[1] );
		
		pState->bCoolFirst[1] = UniformRandomNumber() > 0.5f ? True : False;
        
        // Only allow a cool/warm top/bottom flip if it won't result in a loss of top/bottom contrast midway through the blend
        if ( pState->bCoolFirst[1] != pState->bCoolFirst[0] )
        {
            Color3 vTmpColorA, vTmpColorB;
            LerpColor3( &vTmpColorA, pState->bCoolFirst[0] ? &pState->vCoolColors[0] : &pState->vWarmColors[0], pState->bCoolFirst[1] ? &pState->vCoolColors[1] : &pState->vWarmColors[1], 0.5f );
            LerpColor3( &vTmpColorB, pState->bCoolFirst[0] ? &pState->vWarmColors[0] : &pState->vCoolColors[0], pState->bCoolFirst[1] ? &pState->vWarmColors[1] : &pState->vCoolColors[1], 0.5f );
            
            float fDiff = 0.0f;
            for ( int n=0; n<3; n++ )
            {
                fDiff += fabs(vTmpColorA.RGB[n] - vTmpColorB.RGB[n]);
            }
            
            // If we're below the threashold, then don't do the flip or we'll lose contrast in the gradient
            if ( fDiff <= CONTRAST_THRESHOLD )
            {
                pState->bCoolFirst[1] = pState->bCoolFirst[0]; // Don't flip
            }
        }
	}
	
	float fT = SmoothStep( 0.f, 1.f, (pState->fTimeCounter / fTransitionTime) );
	Color3 vColorA;
	Color3 vColorB;
	
	LerpColor3( &vColorA, pState->bCoolFirst[0] ? &pState->vCoolColors[0] : &pState->vWarmColors[0], pState->bCoolFirst[1] ? &pState->vCoolColors[1] : &pState->vWarmColors[1], fT );
	LerpColor3( &vColorB, pState->bCoolFirst[0] ? &pState->vWarmColors[0] : &pState->vCoolColors[0], pState->bCoolFirst[1] ? &pState->vWarmColors[1] : &pState->vCoolColors[1], fT );
	
	// Bind the gradient shader
	GL_DEBUG( glUseProgramObjectARB( pState->hGradientShader ) );
	
	// Set the colors for the gradient shader
	GL_DEBUG( glUniform4fARB( pState->nGradientColorShaderParamIDs[0], vColorA.r, vColorA.g, vColorA.b, 1.f ) );
	GL_DEBUG( glUniform4fARB( pState->nGradientColorShaderParamIDs[1], vColorB.r, vColorB.g, vColorB.b, 1.f ) );
	
	// Bind the background FBO
	GL_DEBUG( glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, pState->nBackgroundFBO ) );
	GL_DEBUG( glViewport( 0, 0, 1, pState->nRenderTargetHeight ) );
	
	// No need to clear, but I'm doing it anyway
	GL_DEBUG( glClear( GL_COLOR_BUFFER_BIT ) );
	
	// Draw the gradient into the background render target
	GL_DEBUG( DrawScreenQuad( pState, NULL, NULL, NULL ) );
}

// ======================================================================
// Clear the sim and return it to initial state
// ======================================================================
static void ResetSimulation ( PuddleStateStruct* pState )
{
    // Clear sim buffer to all 0.5's (we store packed values there, so 0.5 is unpacked to 0)
	GL_DEBUG( glClearColor( 0.5f, 0.5f, 0.5f, 1.f ) );
	
	int nBuffer;
	for ( nBuffer = 0; nBuffer < 2; nBuffer++ )
	{
		// Bind and clear simulation buffer FBO
		GL_DEBUG( glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, pState->nWaterSimFBOs[nBuffer] ) );
		GL_DEBUG( glViewport( 0, 0, pState->nSimulationResolution, pState->nSimulationResolution ) );
		GL_DEBUG( glClear( GL_COLOR_BUFFER_BIT ) );
	}
	
    // Set the clear color back to what it (probably?) was
	GL_DEBUG( glClearColor( 0.f, 0.f, 0.f, 1.f ) );
    
	// Bind the standard back buffer
	GL_DEBUG( glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, 0 ) );
	GL_DEBUG( glViewport( 0, 0, pState->nWindowWidth, pState->nWindowHeight ) );
    
    // Reset any droplet list
    for ( int nDroplet=0; nDroplet<MAX_DROPLET_COUNT; nDroplet++)
    {
        pState->pDropletList[nDroplet].z = DROPLET_MAX_AGE;
    }
}

// ======================================================================================
// Update the droplet list, each droplet will inject a bit of enegery into the simulation
// ======================================================================================
static void UpdateDroplets ( PuddleStateStruct* pState )
{
    // Probabilistic droplet injection. Use a Poisson random number to figure out how many droplets
    // were generated over the last time interval.
    const float fNumDropletsPerSecond = g_nNumDropletsPerSecond;
    const float fDeltaTime = pState->fTimeStep;
    const float fExpectedDroplets = fNumDropletsPerSecond * fDeltaTime;
    int nNumNewDroplets = PoissonRandomNumber( fExpectedDroplets );
    
    // Add new droplets in any slot where there's a dead droplet
    for ( int nDroplet=0; (nDroplet<MAX_DROPLET_COUNT) && (nNumNewDroplets>0); nDroplet++)
    {
        // Look for dead droplets
        if ( pState->pDropletList[nDroplet].z >= DROPLET_MAX_AGE )
        {
            // Give new droplet a random position and set age to random
            pState->pDropletList[nDroplet].x = UniformRandomNumber() * 2.0f - 1.0f;
            pState->pDropletList[nDroplet].y = UniformRandomNumber() * 2.0f - 1.0f;
            pState->pDropletList[nDroplet].z = UniformRandomNumber() * (0.5f * DROPLET_MAX_AGE);
            
            --nNumNewDroplets;
        }
    }
}

// ======================================================================
// Step the water ripple sim
// ======================================================================
static void UpdateRipples ( PuddleStateStruct* pState )
{
	int nInputBuffer, nOutputBuffer;
	
	// Step the simulation
	for ( int nPass = 0; nPass < 2; nPass++ )
	{
		nInputBuffer = nPass % 2;
		nOutputBuffer = nInputBuffer == 0 ? 1 : 0;
		pState->nNewestSimBufferIndex = nOutputBuffer;
		
		// Bind the simulation shader
		GL_DEBUG( glUseProgramObjectARB( pState->hSimulationShader ) );
		
		// Hook up the sampler and the parameters
		GL_DEBUG( glActiveTexture( GL_TEXTURE0 + 0 ) );
		GL_DEBUG( glBindTexture( GL_TEXTURE_2D, pState->nWaterSimTextures[nInputBuffer] ) );
		GL_DEBUG( glUniform1iARB( pState->nWaterSimSamplerID, 0 ) );
		
		GL_DEBUG( glUniform2fARB( pState->nTextureResolutionParamID, pState->nSimulationResolution, pState->nSimulationResolution ) );
		GL_DEBUG( glUniform1fARB( pState->nTimeStepParamID, pState->fTimeStep ) );
		
		// Bind simulation buffer 0 FBO
		GL_DEBUG( glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, pState->nWaterSimFBOs[nOutputBuffer] ) );
		GL_DEBUG( glViewport( 0, 0, pState->nSimulationResolution, pState->nSimulationResolution ) );
		
		// Update the simulation buffer
		GL_DEBUG( DrawScreenQuad( pState, NULL, NULL, NULL ) );
		
		// Unbind the texture
		GL_DEBUG( glBindTexture( GL_TEXTURE_2D, 0 ) );
	}
    
    // Now inject any living droplets into the sim to add some energy. This is intentionally done
    // after the sim ripple buffer update so that we see a frame of "droplet" each time a new one is spawned
    const float fRainDropSizeScale = 4.f;
    float fRainDropSize = fRainDropSizeScale / (float)pState->nSimulationResolution;
    Vector3 vRainDropScale;
    vRainDropScale.x = fRainDropSize;
    vRainDropScale.y = fRainDropSize;
    vRainDropScale.z = 1.0f;
    
    // Bind simulation buffer 0 FBO (this is always the first "input" buffer each update, which is why nNumSimSteps must always be an even number)
    GL_DEBUG( glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, pState->nWaterSimFBOs[0] ) );
    GL_DEBUG( glViewport( 0, 0, pState->nSimulationResolution, pState->nSimulationResolution ) );
    
    GL_DEBUG( glEnable( GL_BLEND ) );
    GL_DEBUG( glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA ) );
    
    // Bind the rain drop shader
    GL_DEBUG( glUseProgramObjectARB( pState->hRainDropShader ) );
    
    // Inject droplet energy into the simulation
    for ( int nDroplet=0; nDroplet < MAX_DROPLET_COUNT; ++nDroplet )
    {
        // If the droplet is alive, draw it
        if ( pState->pDropletList[nDroplet].z < DROPLET_MAX_AGE )
        {
            const Vector3 vDroplet = pState->pDropletList[nDroplet];
            const float fAge = vDroplet.z / DROPLET_MAX_AGE;
            
            glBegin(GL_QUADS);
            
            glTexCoord4f( 0.0f, 0.0f, 0.0f, fAge );
            glVertex4f( -fRainDropSize + vDroplet.x, -fRainDropSize + vDroplet.y, 0.5f, 1.0f );
            
            glTexCoord4f( 1.0f, 0.0f, 0.0f, fAge );
            glVertex4f(  fRainDropSize + vDroplet.x, -fRainDropSize + vDroplet.y, 0.5f, 1.0f );
            
            glTexCoord4f( 1.0f, 1.0f, 0.0f, fAge );
            glVertex4f(  fRainDropSize + vDroplet.x,  fRainDropSize + vDroplet.y, 0.5f, 1.0f );
            
            glTexCoord4f( 0.0f, 1.0f, 0.0f, fAge );
            glVertex4f( -fRainDropSize + vDroplet.x,  fRainDropSize + vDroplet.y, 0.5f, 1.0f );
            
            glEnd();
            
            // Time passes, the droplet grows older...
            pState->pDropletList[nDroplet].z += pState->fTimeStep;
        }
        
    }
    
    GL_DEBUG( glDisable(GL_BLEND) );
    GL_DEBUG( glUseProgramObjectARB( 0 ) );
}

// ======================================================================
// Update the water surface normals
// ======================================================================
static void UpdateNormalMap ( PuddleStateStruct* pState )
{
	// Bind the normal map generation shader
	GL_DEBUG( glUseProgramObjectARB( pState->hNormalGenShader ) );
	
	// Hook up the sampler and the parameters
	GL_DEBUG( glActiveTexture( GL_TEXTURE0 + 0 ) );
	GL_DEBUG( glBindTexture( GL_TEXTURE_2D, pState->nWaterSimTextures[pState->nNewestSimBufferIndex] ) );
	GL_DEBUG( glUniform1iARB( pState->nNormalGenSamplerID, 0 ) );
	
	GL_DEBUG( glUniform2fARB( pState->nTextureResolutionParamID, pState->nSimulationResolution, pState->nSimulationResolution ) );
    GL_DEBUG( glUniform2fARB( pState->nRippleHeightParamID, g_fRippleHeight, 1.0/g_fRippleHeight ) );
	
	// Bind the normal map FBO
	GL_DEBUG( glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, pState->nNormalFBO ) );
	GL_DEBUG( glViewport( 0, 0, pState->nSimulationResolution, pState->nSimulationResolution ) );
	
	// Generate MIP 0 of the normal map
	GL_DEBUG( DrawScreenQuad( pState, NULL, NULL, NULL ) );
    
	// Unbind the texture
	GL_DEBUG( glBindTexture( GL_TEXTURE_2D, 0 ) );
}

// ======================================================================
// Update the ripple simulation
// ======================================================================
static void UpdateSimulation( PuddleStateStruct* pState )
{
    // Update the droplet list
    UpdateDroplets( pState );
    
    // Step the ripple sim
    UpdateRipples( pState );
    
    // Update the normals with new info from the simulation
    UpdateNormalMap(pState );
}

// ======================================================================
// Debug routine for dumping a movie of the screen saver. Just hacking
// this in, not really meant for serious use. It leaks memory.
// ======================================================================
#if ENABLE_FRAME_DUMPING
static void ProcessMovieFrame( PuddleStateStruct* pState )
{
    static unsigned char* pImageBuffer = NULL;
    static int nFrameCount = 0;
    const int nBufferSize = pState->nWindowWidth * pState->nWindowHeight * 4;
    
    // This leaks memory and also violates XScreenSaver rules about function local static storage!
    if ( pImageBuffer == NULL )
    {
        pImageBuffer = malloc( nBufferSize );
        if ( NULL == pImageBuffer )
        {
            fprintf(stderr, "Ran out of memory!");
            return;
        }
    }
    
    // Get the back buffer's pixels
    glReadPixels( 0, 0, pState->nWindowWidth, pState->nWindowHeight, GL_BGRA, GL_UNSIGNED_BYTE, (GLvoid*)pImageBuffer );
    
    
    // Setup the BMP header
    /*
     struct BMPFileHeader
     {
     char cType[2];     // BMP type, "BM" idetifies MS Windows style bmp
     int nFileSize;     // Size of the BMP file in bytes
     int nReserved;     // Reserved; actual value depends on the application that creates the image
     int nOffset;       // Offset of the byte where the bitmap image data can be found (54)
     }
     
     struct BMPInfoHeader
     {
     int nHeaderSize;   // Size of BMPInfoHeader in bytes (40)
     int nWidth;        // Bitmap width in pixels
     int nHeight;       // Bitmap height in pixels
     short nPlanes;     // Number of color planes being used. Must be set to 1.
     short nNumBits;    // Number of bits per pixel, which is the color depth of the image.
     int nCompression;  // Compression method being used. (0=uncompressed)
     int nImageSize;    // Image size. This is the size of the raw bitmap data, not be confused with the file size.
     int nHPPM;         // Horizontal resolution of the image. (pixel per meter)
     int nVPPM;         // vertical resolution of the image. (pixel per meter)
     int nNumColors;    // Number of colors in the color palette, or 0 to default to 2^n
     int nNumImpColors; // Number of important colors used, or 0 when every color is important; generally ignored.
     };
     
     struct BMPHeader
     {
     BMPFileHeader fileHeader;
     BMPInfoHeader infoHeader;
     };
     
     BMPHeader header;
     header.fileHeader.cType[0] = 'B';
     header.fileHeader.cType[1] = 'M';
     header.fileHeader.nFileSize = sizeof(BMPHeader) + nBufferSize;
     header.fileHeader.nReserved = 0;
     header.fileHeader.nOffset = sizeof(BMPHeader);
     header.infoHeader.nHeaderSize = sizeof(BMPInfoHeader);
     header.infoHeader.nWidth = pState->nWindowWidth;
     header.infoHeader.nHeight = pState->nWindowHeight;
     header.infoHeader.nPlanes = 1;
     header.infoHeader.nNumBits = 32;
     header.infoHeader.nCompression = 0;
     header.infoHeader.nImageSize = nBufferSize;
     header.infoHeader.nHPPM = 0;
     header.infoHeader.nVPPM = 0;
     header.infoHeader.nNumColors = 0;
     header.infoHeader.nNumImpColors = 0;
     */
    
    unsigned char pBMPFileHeader[14] = {'B','M', 0,0,0,0, 0,0, 0,0, 54,0,0,0};
    unsigned char pBMPInfoHeader[40] = {40,0,0,0, 0,0,0,0, 0,0,0,0, 1,0, 32,0};
    
    *(int*)(pBMPFileHeader+2)  = sizeof(pBMPFileHeader) + sizeof(pBMPInfoHeader) + nBufferSize;
    *(int*)(pBMPInfoHeader+4)  = pState->nWindowWidth;
    *(int*)(pBMPInfoHeader+8)  = pState->nWindowHeight;
    *(int*)(pBMPInfoHeader+20) = pState->nWindowWidth * pState->nWindowHeight * 4;
    
    
    // Make a file name
    char strFileName[1024] = { '\0' };
    sprintf( &strFileName[0], "%s/frame_%.5d.bmp", FRAME_DUMP_LOCATION, nFrameCount++ );
    
    // Open the file for writing
    FILE* pFile = fopen( strFileName, "wb" );
    if ( NULL == pFile )
    {
        fprintf(stderr, "Unable to open file for writing: %s\n", strFileName);
        return;
    }
    
    // Write the header & pixels
    fwrite( pBMPFileHeader, 1, sizeof(pBMPFileHeader), pFile );
    fwrite( pBMPInfoHeader, 1, sizeof(pBMPInfoHeader), pFile );
    fwrite( pImageBuffer, pState->nWindowWidth * pState->nWindowHeight * 4, 1, pFile );
    
    fclose( pFile );
    
}
#endif // ENABLE_FRAME_DUMPING

// ======================================================================
// Ripple screen saver rendering entry point
// ======================================================================
static void RippleDisplay ( PuddleStateStruct* pState )
{
	GL_DEBUG_LABEL( "Entering RippleDisplay()" );
	
    GL_DEBUG( glDisable( GL_LIGHTING ) );
	GL_DEBUG( glDisable( GL_DEPTH_TEST ) );
	GL_DEBUG( glDisable( GL_CULL_FACE ) );
    
    GL_DEBUG( glClearColor(0.f, 0.f, 0.f, 1.f) );
    //GL_DEBUG( glClearDepthf( 0.0f ) );
    //GL_DEBUG( glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT) );
    
	
	// Compute new gradient colors
    if ( !pState->bPauseGradient )
    {
        UpdateColorGradient( pState );
    }
    
	// Reset simulation if requested
	if ( pState->bResetSim )
	{
		ResetSimulation( pState );
		pState->bResetSim = FALSE;
	}
	
	// Update simulation
	if ( !pState->bPauseSim )
	{
		UpdateSimulation(pState );
	}
	
	// Bind the standard back buffer
	GL_DEBUG( glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, 0 ) );
	GL_DEBUG( glViewport( 0, 0, pState->nWindowWidth, pState->nWindowHeight ) );
	
	// No need to clear, but I'm doing it anyway
	GL_DEBUG( glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT ) );
    
	// Draw water ripple effect
	GL_DEBUG( DrawRipplePlane( pState ) );
    
    // Debug visualization
	DrawDebugVisualization( pState );
	
	// Disable the programmable pipeline and return to fixed function execution
	GL_DEBUG( glUseProgramObjectARB( 0 ) );
    
    // The screensaver framework seems to require that we leave this state active or it won't
    // render text (frame rate, CPU load, etc) correctly.
    GL_DEBUG( glActiveTexture( GL_TEXTURE0 + 0 ) );
    
#if ENABLE_FRAME_DUMPING
    if ( pState->bDumpMovieFrames || pState->bDumpScreenShot )
    {
        GL_DEBUG( ProcessMovieFrame( pState ) );
        pState->bDumpScreenShot = FALSE;
    }
#endif // ENABLE_FRAME_DUMPING
}


/*
 *-----------------------------------------------------------------------------
 *-----------------------------------------------------------------------------
 *    Xlock hooks.
 *-----------------------------------------------------------------------------
 *-----------------------------------------------------------------------------
 */



// *-----------------------------------------------------------------------------
// *
// *-----------------------------------------------------------------------------
ENTRYPOINT void reshape_puddles ( ModeInfo * mi, int width, int height )
{
    //DEBUG_LABEL( "reshape_puddles" );
    
	PuddleStateStruct *pState = &g_pPuddleState[MI_SCREEN(mi)];
    
	glViewport(0, 0, pState->nWindowWidth = (GLint) width, pState->nWindowHeight = (GLint) height);
	
	float mProjMatrix[16];
	float mModelViewMatrix[16];
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective( 45.0f, (GLdouble) pState->nWindowWidth / (GLdouble) pState->nWindowHeight, 0.1f, 100.0 );
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	gluLookAt( 0.0f, -1.0f, 0.125f, 0.0f, 1.0f, -1.0f, 0.0f, 1.0f, 0.0f );
	
	glGetFloatv( GL_MODELVIEW_MATRIX, mModelViewMatrix );
	glGetFloatv( GL_PROJECTION_MATRIX, mProjMatrix );
	
	// Create composite MVP matrix
	pState->mModelViewProjMatrix[0]  = mModelViewMatrix[0]*mProjMatrix[0]  + mModelViewMatrix[1]*mProjMatrix[4]  + mModelViewMatrix[2]*mProjMatrix[8]  + mModelViewMatrix[3]*mProjMatrix[12];
	pState->mModelViewProjMatrix[1]  = mModelViewMatrix[0]*mProjMatrix[1]  + mModelViewMatrix[1]*mProjMatrix[5]  + mModelViewMatrix[2]*mProjMatrix[9]  + mModelViewMatrix[3]*mProjMatrix[13];
	pState->mModelViewProjMatrix[2]  = mModelViewMatrix[0]*mProjMatrix[2]  + mModelViewMatrix[1]*mProjMatrix[6]	 + mModelViewMatrix[2]*mProjMatrix[10] + mModelViewMatrix[3]*mProjMatrix[14];
	pState->mModelViewProjMatrix[3]  = mModelViewMatrix[0]*mProjMatrix[3]  + mModelViewMatrix[1]*mProjMatrix[7]  + mModelViewMatrix[2]*mProjMatrix[11] + mModelViewMatrix[3]*mProjMatrix[15];
	
	pState->mModelViewProjMatrix[4]  = mModelViewMatrix[4]*mProjMatrix[0]  + mModelViewMatrix[5]*mProjMatrix[4]  + mModelViewMatrix[6]*mProjMatrix[8]  + mModelViewMatrix[7]*mProjMatrix[12];
	pState->mModelViewProjMatrix[5]  = mModelViewMatrix[4]*mProjMatrix[1]  + mModelViewMatrix[5]*mProjMatrix[5]  + mModelViewMatrix[6]*mProjMatrix[9]  + mModelViewMatrix[7]*mProjMatrix[13];
	pState->mModelViewProjMatrix[6]  = mModelViewMatrix[4]*mProjMatrix[2]  + mModelViewMatrix[5]*mProjMatrix[6]  + mModelViewMatrix[6]*mProjMatrix[10] + mModelViewMatrix[7]*mProjMatrix[14];
	pState->mModelViewProjMatrix[7]  = mModelViewMatrix[4]*mProjMatrix[3]  + mModelViewMatrix[5]*mProjMatrix[7]  + mModelViewMatrix[6]*mProjMatrix[11] + mModelViewMatrix[7]*mProjMatrix[15];
	
	pState->mModelViewProjMatrix[8]  = mModelViewMatrix[8]*mProjMatrix[0]  + mModelViewMatrix[9]*mProjMatrix[4]  + mModelViewMatrix[10]*mProjMatrix[8]  + mModelViewMatrix[11]*mProjMatrix[12];
	pState->mModelViewProjMatrix[9]  = mModelViewMatrix[8]*mProjMatrix[1]  + mModelViewMatrix[9]*mProjMatrix[5]  + mModelViewMatrix[10]*mProjMatrix[9]  + mModelViewMatrix[11]*mProjMatrix[13];
	pState->mModelViewProjMatrix[10] = mModelViewMatrix[8]*mProjMatrix[2]  + mModelViewMatrix[9]*mProjMatrix[6]  + mModelViewMatrix[10]*mProjMatrix[10] + mModelViewMatrix[11]*mProjMatrix[14];
	pState->mModelViewProjMatrix[11] = mModelViewMatrix[8]*mProjMatrix[3]  + mModelViewMatrix[9]*mProjMatrix[7]  + mModelViewMatrix[10]*mProjMatrix[11] + mModelViewMatrix[11]*mProjMatrix[15];
	
	pState->mModelViewProjMatrix[12] = mModelViewMatrix[12]*mProjMatrix[0] + mModelViewMatrix[13]*mProjMatrix[4] + mModelViewMatrix[14]*mProjMatrix[8]  + mModelViewMatrix[15]*mProjMatrix[12];
	pState->mModelViewProjMatrix[13] = mModelViewMatrix[12]*mProjMatrix[1] + mModelViewMatrix[13]*mProjMatrix[5] + mModelViewMatrix[14]*mProjMatrix[9]  + mModelViewMatrix[15]*mProjMatrix[13];
	pState->mModelViewProjMatrix[14] = mModelViewMatrix[12]*mProjMatrix[2] + mModelViewMatrix[13]*mProjMatrix[6] + mModelViewMatrix[14]*mProjMatrix[10] + mModelViewMatrix[15]*mProjMatrix[14];
	pState->mModelViewProjMatrix[15] = mModelViewMatrix[12]*mProjMatrix[3] + mModelViewMatrix[13]*mProjMatrix[7] + mModelViewMatrix[14]*mProjMatrix[11] + mModelViewMatrix[15]*mProjMatrix[15];
    
}


// *-----------------------------------------------------------------------------
// *    Initialize puddles screen saver.  Called each time the window changes.
// *-----------------------------------------------------------------------------
ENTRYPOINT void init_puddles ( ModeInfo* pMI )
{
    //DEBUG_LABEL( "init_puddles" );
    
	int screen = MI_SCREEN(pMI);
	PuddleStateStruct* pState;
	Display* pDisplay = MI_DISPLAY(pMI);
	Window window = MI_WINDOW(pMI);
    
	if (g_pPuddleState == NULL)
	{
		if ( (g_pPuddleState = (PuddleStateStruct *) calloc(MI_NUM_SCREENS(pMI), sizeof (PuddleStateStruct))) == NULL)
        {
			return;
        }
	}
	
	pState = &g_pPuddleState[screen];
    
    
	if ( MI_IS_DEBUG(pMI) )
	{
		(void) fprintf(stderr, "%s\n", MI_NAME(pMI) );
	}
	
	if ( ( pState->pGLXContext = init_GL(pMI) ) != NULL )
	{
		reshape_puddles(pMI, MI_WIDTH(pMI), MI_HEIGHT(pMI));
		glDrawBuffer(GL_BACK);
		Init(pMI);
		RippleDisplay(pState);
		glXSwapBuffers(pDisplay, window);
        
	}
	else
	{
		MI_CLEARWINDOW(pMI);
	}
}

/*
 *-----------------------------------------------------------------------------
 *    Event handler (doesn't do anything)
 *-----------------------------------------------------------------------------
 */
ENTRYPOINT Bool puddles_handle_event ( ModeInfo *pMI, XEvent *pEvent )
{
    //DEBUG_LABEL( "puddles_handle_event" );
    
	int screen = MI_SCREEN(pMI);
	PuddleStateStruct* pState = &g_pPuddleState[screen];
	
	if ( pEvent->xany.type == KeyPress )
	{
		switch ( XKeycodeToKeysym(pMI->dpy, pEvent->xkey.keycode, 0) )
		{
			case 'r' :
			case 'R' :
				pState->bResetSim = TRUE;
				return True;
				
			case 'p' :
				pState->bPauseSim = !pState->bPauseSim;
				return True;
                
            case 'P' :
				pState->bPauseGradient = !pState->bPauseGradient;
				return True;
                
			case 't' :
				pState->nDebugTextureIndex++;
				return True;
				
			case 'T' :
				pState->nDebugTextureIndex--;
				return True;
				
			case 'g' :
			case 'G' :
				pState->bRegenerateGradient = TRUE;
				return True;
                
#if ENABLE_FRAME_DUMPING
            case 's' :
            case 'S' :
                pState->bDumpScreenShot = TRUE;
                return True;
#endif // ENABLE_FRAME_DUMPING

		}
	}
	
	
	return False;
}

/*
 *-----------------------------------------------------------------------------
 *    Called by the mainline code periodically to update the display.
 *-----------------------------------------------------------------------------
 */
ENTRYPOINT void draw_puddles ( ModeInfo* pMI )
{
    //DEBUG_LABEL( "draw_puddles" );
    
	PuddleStateStruct* pState = &g_pPuddleState[MI_SCREEN(pMI)];
    
	Display* pDisplay = MI_DISPLAY(pMI);
	Window window = MI_WINDOW(pMI);
    
	MI_IS_DRAWN(pMI) = True;
	
	// TODO: Make sure this assumption is legit...
	// Assume that we're updating at 30 Hz
	pState->fTimeStep = 1.f / 30.f;
	pState->fTimeCounter += pState->fTimeStep;
    
	if ( !pState->pGLXContext )
		return;
    
	glXMakeCurrent(pDisplay, window, *(pState->pGLXContext));
    
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
	RippleDisplay(pState);
    
	if (pMI->fps_p) do_fps (pMI);
	{
		glXSwapBuffers(pDisplay, window);
	}
}


/*
 *-----------------------------------------------------------------------------
 *    The display is being taken away from us.  Free up malloc'ed
 *      memory and X resources that we've alloc'ed.  Only called
 *      once, we must zap everything for every screen.
 *-----------------------------------------------------------------------------
 */
ENTRYPOINT void release_puddles(ModeInfo* pMI)
{
    //DEBUG_LABEL( "release_puddles" );
    
	int screen;
    
	if ( g_pPuddleState != NULL )
	{
		for (screen = 0; screen < MI_NUM_SCREENS(pMI); screen++)
		{
			PuddleStateStruct* pState = &g_pPuddleState[screen];
            
			// free any per-screen allocations
			
			// Delete shaders
			GL_DEBUG( glDeleteObjectARB( pState->hGradientShader ) );
			GL_DEBUG( glDeleteObjectARB( pState->hWaterRippleShader ) );
			GL_DEBUG( glDeleteObjectARB( pState->hPassThroughShader ) );
			GL_DEBUG( glDeleteObjectARB( pState->hRainDropShader ) );
			GL_DEBUG( glDeleteObjectARB( pState->hSimulationShader ) );
			GL_DEBUG( glDeleteObjectARB( pState->hNormalGenShader ) );
            
			GL_DEBUG( glDeleteObjectARB( pState->hGradientFS ) );
			GL_DEBUG( glDeleteObjectARB( pState->hWaterRippleVS ) );
			GL_DEBUG( glDeleteObjectARB( pState->hWaterRippleFS ) );
			GL_DEBUG( glDeleteObjectARB( pState->hPassThroughFS ) );
			GL_DEBUG( glDeleteObjectARB( pState->hRainDropFS ) );
			GL_DEBUG( glDeleteObjectARB( pState->hSimulationFS ) );
			GL_DEBUG( glDeleteObjectARB( pState->hNormalGenFS ) );
			
			GL_DEBUG( glDeleteObjectARB( pState->hPassThroughVS ) );
			GL_DEBUG( glDeleteObjectARB( pState->hSimulationVS ) );
			
			// Delete frame buffer objects
			GL_DEBUG( glDeleteFramebuffersEXT( 1, &pState->nNormalFBO ) );
			GL_DEBUG( glDeleteFramebuffersEXT( 2, &pState->nWaterSimFBOs[0] ) );
			GL_DEBUG( glDeleteFramebuffersEXT( 1, &pState->nBackgroundFBO ) );
			
			// Delete the textures
			GL_DEBUG( glDeleteTextures( 1, &pState->nNormalTexture ) );
			GL_DEBUG( glDeleteTextures( 2, &pState->nWaterSimTextures[0] ) );
			GL_DEBUG( glDeleteTextures( 1, &pState->nBackgroundTexture ) );
			
		}
		free((void *) g_pPuddleState);
		g_pPuddleState = NULL;
	}
	FreeAllGL(pMI);
}

//==============================================================================
//
//==============================================================================
ENTRYPOINT void refresh_puddles (ModeInfo* pMI)
{
    //DEBUG_LABEL( "refresh_puddles" );
}

//==============================================================================
//
//==============================================================================
#ifndef STANDALONE
ENTRYPOINT void change_puddles (ModeInfo* pMI)
{
    //DEBUG_LABEL( "change_puddles" );
    
	PuddleStateStruct* pState = &puddles[MI_SCREEN(pMI)];
    
	if (!pState->glx_context)
		return;
    
	glXMakeCurrent(MI_DISPLAY(pMI), MI_WINDOW(pMI), *(pState->glx_context));
	Init(mi);
}
#endif /* !STANDALONE */

XSCREENSAVER_MODULE ("Puddles", puddles)

#endif /* USE_GL */
